# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import spectre
from gnuradio import soapy

import spectre_core.fields

from ._base import Base, BaseModel
from ._constants import FlowgraphConstant


class HackRFFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2e6
    bandwidth: spectre_core.fields.Field.bandwidth = 2e6
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    amp_on: spectre_core.fields.Field.amp_on = False
    lna_gain: spectre_core.fields.Field.lna_gain = 20
    vga_gain: spectre_core.fields.Field.vga_gain = 20
    batch_size: spectre_core.fields.Field.batch_size = 3
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class HackRFFixedCenterFrequency(Base[HackRFFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: HackRFFixedCenterFrequencyModel) -> None:
        device = "driver=hackrf"
        type = model.output_type
        nchan = 1
        dev_args = ""
        stream_args = ""
        tune_args = [""]
        settings = [""]
        self.soapy_hackrf_source = soapy.source(
            device, type, nchan, dev_args, stream_args, tune_args, settings
        )
        self.soapy_hackrf_source.set_sample_rate(0, model.sample_rate)
        self.soapy_hackrf_source.set_bandwidth(0, model.bandwidth)
        self.soapy_hackrf_source.set_frequency(0, model.center_frequency)
        self.soapy_hackrf_source.set_gain(0, "AMP", model.amp_on)
        self.soapy_hackrf_source.set_gain(0, "LNA", model.lna_gain)
        self.soapy_hackrf_source.set_gain(0, "VGA", model.vga_gain)

        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
        )
        self.connect((self.soapy_hackrf_source, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

"""Shared pydantic field values."""

from ._fields import Field
from ._field_values import OutputType, WindowType

__all__ = ["Field", "OutputType", "WindowType"]



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import time
import dataclasses

from gnuradio import spectre
from gnuradio import uhd

import spectre_core.fields

from ._base import Base, BaseModel
from ._constants import FlowgraphConstant


@dataclasses.dataclass(frozen=True)
class USRPWireFormat:
    """Indicates the form of the data over the bus/network."""

    SC8 = "sc8"
    SC12 = "sc12"
    SC16 = "sc16"


class USRPFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 600000
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 600000
    gain: spectre_core.fields.Field.gain = 35
    wire_format: spectre_core.fields.Field.wire_format = USRPWireFormat.SC12
    master_clock_rate: spectre_core.fields.Field.master_clock_rate = 60000000
    num_recv_frames: spectre_core.fields.Field.num_recv_frames = 32
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class USRPFixedCenterFrequency(Base[USRPFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: USRPFixedCenterFrequencyModel) -> None:
        master_clock_rate = f"master_clock_rate={model.master_clock_rate}"
        num_recv_frames = f"num_recv_frames={model.num_recv_frames}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate, num_recv_frames)),
            uhd.stream_args(
                cpu_format=model.output_type,
                otw_format=model.wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(model.sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source.set_center_freq(model.center_frequency, 0)
        self.uhd_usrp_source.set_bandwidth(model.bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(model.gain, 0)
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
        )
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))


class USRPSweptCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2000000
    batch_size: spectre_core.fields.Field.batch_size = 3
    bandwidth: spectre_core.fields.Field.bandwidth = 2e6
    min_frequency: spectre_core.fields.Field.min_frequency = 95e6
    max_frequency: spectre_core.fields.Field.max_frequency = 101e6
    gain: spectre_core.fields.Field.gain = 35
    wire_format: spectre_core.fields.Field.wire_format = USRPWireFormat.SC12
    master_clock_rate: spectre_core.fields.Field.master_clock_rate = 60000000
    num_recv_frames: spectre_core.fields.Field.num_recv_frames = 32
    dwell_time: spectre_core.fields.Field.dwell_time = 0.15
    frequency_hop: spectre_core.fields.Field.frequency_hop = 2e6
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class USRPSweptCenterFrequency(Base[USRPSweptCenterFrequencyModel]):
    def configure(self, tag: str, model: USRPSweptCenterFrequencyModel) -> None:
        master_clock_rate = f"master_clock_rate={model.master_clock_rate}"
        num_recv_frames = f"num_recv_frames={model.num_recv_frames}"
        self.uhd_usrp_source = uhd.usrp_source(
            ",".join(("", "", master_clock_rate, num_recv_frames)),
            uhd.stream_args(
                cpu_format=model.output_type,
                otw_format=model.wire_format,
                args="",
                channels=[0],
            ),
        )
        self.uhd_usrp_source.set_samp_rate(model.sample_rate)
        self.uhd_usrp_source.set_time_now(uhd.time_spec(time.time()), uhd.ALL_MBOARDS)
        self.uhd_usrp_source.set_center_freq(model.min_frequency, 0)
        self.uhd_usrp_source.set_bandwidth(model.bandwidth, 0)
        self.uhd_usrp_source.set_rx_agc(False, 0)
        self.uhd_usrp_source.set_auto_dc_offset(False, 0)
        self.uhd_usrp_source.set_auto_iq_balance(False, 0)
        self.uhd_usrp_source.set_gain(model.gain, 0)

        retune_cmd_name = "freq"
        self.spectre_frequency_sweeper = spectre.frequency_sweeper(
            model.min_frequency,
            model.max_frequency,
            model.frequency_hop,
            model.dwell_time,
            model.sample_rate,
            retune_cmd_name,
            model.output_type,
        )

        is_tagged = True
        frequency_tag_key = "rx_freq"
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
            is_tagged,
            frequency_tag_key,
            model.min_frequency,
        )

        self.msg_connect(
            (self.spectre_frequency_sweeper, "retune_command"),
            (self.uhd_usrp_source, "command"),
        )
        self.connect((self.uhd_usrp_source, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.uhd_usrp_source, 0), (self.spectre_frequency_sweeper, 0))



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import gr
from gnuradio import blocks
from gnuradio import analog
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel
from ._constants import FlowgraphConstant


class SignalGeneratorCosineWaveModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 128e3
    batch_size: spectre_core.fields.Field.batch_size = 1
    frequency: spectre_core.fields.Field.frequency = 32e3
    amplitude: spectre_core.fields.Field.amplitude = 1
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class SignalGeneratorCosineWave(Base[SignalGeneratorCosineWaveModel]):
    def configure(self, tag: str, model: SignalGeneratorCosineWaveModel) -> None:
        """Record a complex-valued cosine signal."""
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
        )
        self.blocks_throttle_0 = blocks.throttle(
            gr.sizeof_float * 1, model.sample_rate, True
        )
        self.blocks_throttle_1 = blocks.throttle(
            gr.sizeof_float * 1, model.sample_rate, True
        )
        self.blocks_null_source = blocks.null_source(gr.sizeof_float * 1)
        self.blocks_float_to_complex = blocks.float_to_complex(1)

        self.analog_sig_source = analog.sig_source_f(
            model.sample_rate,
            analog.GR_COS_WAVE,
            model.frequency,
            model.amplitude,
        )

        self.connect((self.analog_sig_source, 0), (self.blocks_throttle_0, 0))
        self.connect((self.blocks_null_source, 0), (self.blocks_throttle_1, 0))
        self.connect((self.blocks_throttle_0, 0), (self.blocks_float_to_complex, 0))
        self.connect((self.blocks_throttle_1, 0), (self.blocks_float_to_complex, 1))
        self.connect(
            (self.blocks_float_to_complex, 0), (self.spectre_batched_file_sink, 0)
        )


class SignalGeneratorConstantStaircaseModel(BaseModel):
    step_increment: spectre_core.fields.Field.step_increment = 200
    sample_rate: spectre_core.fields.Field.sample_rate = 128000
    min_samples_per_step: spectre_core.fields.Field.min_samples_per_step = 4000
    max_samples_per_step: spectre_core.fields.Field.max_samples_per_step = 5000
    frequency_hop: spectre_core.fields.Field.frequency_hop = 128000
    batch_size: spectre_core.fields.Field.batch_size = 1
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class SignalGeneratorConstantStaircase(Base[SignalGeneratorConstantStaircaseModel]):
    def configure(self, tag: str, model: SignalGeneratorConstantStaircaseModel) -> None:
        """Record a constant signal that periodically increments in value.
        Each step increases in duration, up to a maximum, before resetting."""
        self.spectre_constant_staircase = spectre.tagged_staircase(
            model.min_samples_per_step,
            model.max_samples_per_step,
            model.frequency_hop,
            model.step_increment,
            model.sample_rate,
        )

        is_tagged = True
        frequency_tag_key = "rx_freq"
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
            is_tagged,
            frequency_tag_key,
            FlowgraphConstant.NO_INITIAL_TAG_VALUE,
        )
        self.blocks_throttle = blocks.throttle(
            gr.sizeof_gr_complex * 1, model.sample_rate
        )

        self.connect((self.spectre_constant_staircase, 0), (self.blocks_throttle, 0))
        self.connect((self.blocks_throttle, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses


@dataclasses.dataclass(frozen=True)
class OutputType:
    """The type of samples produced by a source block."""

    FC32 = "fc32"
    FC64 = "fc64"
    SC8 = "sc8"
    SC16 = "sc16"


@dataclasses.dataclass(frozen=True)
class WindowType:
    HANN = "hann"
    BLACKMAN = "blackman"
    BOXCAR = "boxcar"



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses

from gnuradio import sdrplay3
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel
from ._constants import FlowgraphConstant


@dataclasses.dataclass(frozen=True)
class RSPduoPort:
    """Specifies one of the antenna ports on the RSPduo.

    These are easier to type in the CLI tool, than the values we must pass to `gr-sdrplay3`.
    Namely 'Tuner 1 50' ohm and 'Tuner 2 50 ohm'.
    """

    TUNER_1 = "tuner_1"
    TUNER_2 = "tuner_2"


def _map_port(antenna_port: str | None) -> str:
    """Maps the CLI-typeable port, to the one used in the constructor for the RSPduo block in the gr-sdrplay3 OOT module."""
    if antenna_port == RSPduoPort.TUNER_1:
        return "Tuner 1 50 ohm"
    elif antenna_port == RSPduoPort.TUNER_2:
        return "Tuner 2 50 ohm"
    else:
        raise ValueError(f"{antenna_port} is not a valid antenna port.")


class RSPduoFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2e6
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 1.536e6
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )
    antenna_port: spectre_core.fields.Field.antenna_port = RSPduoPort.TUNER_1


class RSPduoFixedCenterFrequency(Base[RSPduoFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: RSPduoFixedCenterFrequencyModel) -> None:
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
        )
        self.sdrplay3_rspduo = sdrplay3.rspduo(
            "",
            rspduo_mode="Single Tuner",
            stream_args=sdrplay3.stream_args(
                output_type=model.output_type, channels_size=1
            ),
        )
        self.sdrplay3_rspduo.set_sample_rate(model.sample_rate)
        self.sdrplay3_rspduo.set_center_freq(model.center_frequency)
        self.sdrplay3_rspduo.set_bandwidth(model.bandwidth)
        self.sdrplay3_rspduo.set_antenna(_map_port(model.antenna_port))
        self.sdrplay3_rspduo.set_gain_mode(False)
        self.sdrplay3_rspduo.set_gain(model.if_gain, "IF")
        self.sdrplay3_rspduo.set_gain(model.rf_gain, "RF")
        self.sdrplay3_rspduo.set_freq_corr(0)
        self.sdrplay3_rspduo.set_dc_offset_mode(False)
        self.sdrplay3_rspduo.set_iq_balance_mode(False)
        self.sdrplay3_rspduo.set_agc_setpoint(-30)
        self.sdrplay3_rspduo.set_rf_notch_filter(False)
        self.sdrplay3_rspduo.set_dab_notch_filter(False)
        self.sdrplay3_rspduo.set_am_notch_filter(False)
        self.sdrplay3_rspduo.set_biasT(False)
        self.sdrplay3_rspduo.set_debug_mode(False)
        self.sdrplay3_rspduo.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo.set_show_gain_changes(False)

        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_batched_file_sink, 0))


class RSPduoSweptCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2e6
    batch_size: spectre_core.fields.Field.batch_size = 3
    bandwidth: spectre_core.fields.Field.bandwidth = 1.536e6
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    min_frequency: spectre_core.fields.Field.min_frequency = 95e6
    max_frequency: spectre_core.fields.Field.max_frequency = 100e6
    dwell_time: spectre_core.fields.Field.dwell_time = 0.15
    frequency_hop: spectre_core.fields.Field.frequency_hop = 2e6
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )
    antenna_port: spectre_core.fields.Field.antenna_port = RSPduoPort.TUNER_1


class RSPduoSweptCenterFrequency(Base[RSPduoSweptCenterFrequencyModel]):
    def configure(self, tag: str, model: RSPduoSweptCenterFrequencyModel) -> None:
        retune_cmd_name = "freq"
        self.spectre_frequency_sweeper = spectre.frequency_sweeper(
            model.min_frequency,
            model.max_frequency,
            model.frequency_hop,
            model.dwell_time,
            model.sample_rate,
            retune_cmd_name,
            model.output_type,
        )

        is_tagged = True
        frequency_tag_key = "freq"
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
            is_tagged,
            frequency_tag_key,
            model.min_frequency,
        )
        self.sdrplay3_rspduo = sdrplay3.rspduo(
            "",
            rspduo_mode="Single Tuner",
            stream_args=sdrplay3.stream_args(
                output_type=model.output_type, channels_size=1
            ),
        )
        self.sdrplay3_rspduo.set_sample_rate(model.sample_rate, True)
        self.sdrplay3_rspduo.set_center_freq(model.min_frequency, True)
        self.sdrplay3_rspduo.set_bandwidth(model.bandwidth)
        self.sdrplay3_rspduo.set_antenna(_map_port(model.antenna_port))
        self.sdrplay3_rspduo.set_gain_mode(False)
        self.sdrplay3_rspduo.set_gain(model.if_gain, "IF", True)
        self.sdrplay3_rspduo.set_gain(model.rf_gain, "RF", True)
        self.sdrplay3_rspduo.set_freq_corr(0)
        self.sdrplay3_rspduo.set_dc_offset_mode(False)
        self.sdrplay3_rspduo.set_iq_balance_mode(False)
        self.sdrplay3_rspduo.set_agc_setpoint(-30)
        self.sdrplay3_rspduo.set_rf_notch_filter(False)
        self.sdrplay3_rspduo.set_dab_notch_filter(False)
        self.sdrplay3_rspduo.set_am_notch_filter(False)
        self.sdrplay3_rspduo.set_biasT(False)
        self.sdrplay3_rspduo.set_stream_tags(True)
        self.sdrplay3_rspduo.set_debug_mode(False)
        self.sdrplay3_rspduo.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rspduo.set_show_gain_changes(False)

        self.msg_connect(
            (self.spectre_frequency_sweeper, "retune_command"),
            (self.sdrplay3_rspduo, "command"),
        )
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_batched_file_sink, 0))
        self.connect((self.sdrplay3_rspduo, 0), (self.spectre_frequency_sweeper, 0))



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass


@dataclass(frozen=True)
class TimeFormat:
    """Package-wide datetime formats.

    :ivar DATE: Format for dates (e.g., '2025-01-11').
    :ivar TIME: Format for times (e.g., '23:59:59').
    :ivar FRACTIONAL_TIME: Format for times with microsecond precision (e.g., '23:59:59.123456')
    :ivar DATETIME: Format for datetimes compliant with the ISO-8601 standard (e.g., '2025-01-11T23:59:59.123456Z')
    """

    DATE = "%Y-%m-%d"
    TIME = "%H:%M:%S"
    FRACTIONAL_TIME = "%H:%M:%S.%f"
    DATETIME = f"{DATE}T{FRACTIONAL_TIME}Z"



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses


@dataclasses.dataclass(frozen=True)
class FlowgraphConstant:
    """Constants across all flowgraphs."""

    GROUP_BY_DATE = True
    NO_INITIAL_TAG_VALUE = 0



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

from gnuradio import sdrplay3
from gnuradio import spectre

import spectre_core.fields

from ._base import Base, BaseModel
from ._constants import FlowgraphConstant


class RSP1BFixedCenterFrequencyModel(BaseModel):
    sample_rate: spectre_core.fields.Field.sample_rate = 2e6
    batch_size: spectre_core.fields.Field.batch_size = 3
    center_frequency: spectre_core.fields.Field.center_frequency = 95.8e6
    bandwidth: spectre_core.fields.Field.bandwidth = 1.536e6
    if_gain: spectre_core.fields.Field.if_gain = -30
    rf_gain: spectre_core.fields.Field.rf_gain = 0
    output_type: spectre_core.fields.Field.output_type = (
        spectre_core.fields.OutputType.FC32
    )


class RSP1BFixedCenterFrequency(Base[RSP1BFixedCenterFrequencyModel]):
    def configure(self, tag: str, model: RSP1BFixedCenterFrequencyModel) -> None:
        self.spectre_batched_file_sink = spectre.batched_file_sink(
            self._batches_dir_path,
            tag,
            model.output_type,
            model.batch_size,
            model.sample_rate,
            FlowgraphConstant.GROUP_BY_DATE,
        )
        self.sdrplay3_rsp1b = sdrplay3.rsp1b(
            "",
            stream_args=sdrplay3.stream_args(
                output_type=model.output_type, channels_size=1
            ),
        )
        self.sdrplay3_rsp1b.set_sample_rate(model.sample_rate)
        self.sdrplay3_rsp1b.set_center_freq(model.center_frequency)
        self.sdrplay3_rsp1b.set_bandwidth(model.bandwidth)
        self.sdrplay3_rsp1b.set_gain_mode(False)
        self.sdrplay3_rsp1b.set_gain(model.if_gain, "IF")
        self.sdrplay3_rsp1b.set_gain(model.rf_gain, "RF")
        self.sdrplay3_rsp1b.set_freq_corr(0)
        self.sdrplay3_rsp1b.set_dc_offset_mode(False)
        self.sdrplay3_rsp1b.set_iq_balance_mode(False)
        self.sdrplay3_rsp1b.set_agc_setpoint(-30)
        self.sdrplay3_rsp1b.set_rf_notch_filter(False)
        self.sdrplay3_rsp1b.set_dab_notch_filter(False)
        self.sdrplay3_rsp1b.set_biasT(False)
        self.sdrplay3_rsp1b.set_debug_mode(False)
        self.sdrplay3_rsp1b.set_sample_sequence_gaps_check(False)
        self.sdrplay3_rsp1b.set_show_gain_changes(False)

        self.connect((self.sdrplay3_rsp1b, 0), (self.spectre_batched_file_sink, 0))



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import dataclasses
import typing

import numpy as np
import numpy.typing as npt
import astropy.io

import spectre_core.config
import spectre_core.spectrograms

from ._base import Base, BatchFile


@dataclasses.dataclass(frozen=True)
class IQStreamBatchExtension:
    """Supported extensions for a `IQStreamBatch`.

    :ivar FITS: Corresponds to the `.fits` file extension.
    :ivar FC32: Corresponds to the `.fc32` file extension.
    :ivar FC64: Corresponds to the `.fc64` file extension.
    :ivar SC8: Corresponds to the `.sc8` file extension.
    :ivar SC16: Corresponds to the `.sc16` file extension.
    :ivar HDR: Corresponds to the `.hdr` file extension.
    """

    FITS: str = "fits"
    FC32: str = "fc32"
    FC64: str = "fc64"
    SC8: str = "sc8"
    SC16: str = "sc16"
    HDR: str = "hdr"


class _Fc32File(BatchFile[npt.NDArray[np.complex64]]):
    def read(self) -> npt.NDArray[np.complex64]:
        """Read single-precision complex, interleaved I/Q samples in the binary format.

        :return: 64-bit complex IQ samples.
        """
        return np.fromfile(self.file_path, dtype=np.complex64)


class _Fc64File(BatchFile[npt.NDArray[np.complex64]]):
    def read(self) -> npt.NDArray[np.complex64]:
        """Read double-precision complex, interleaved I/Q samples in the binary format. A narrowing
        conversion is applied.

        :return: 64-bit complex IQ samples.
        """
        return np.fromfile(self.file_path, dtype=np.complex128).astype(np.complex64)


class _Sc8File(BatchFile[npt.NDArray[np.complex64]]):
    def read(self) -> npt.NDArray[np.complex64]:
        """Read single-precision complex, interleaved I/Q samples in the binary format. A promotional
        conversion is applied.

        :return: 64-bit complex IQ samples.
        """
        data = np.fromfile(self.file_path, dtype=np.int8)
        return (data[0::2] + 1j * data[1::2]).astype(np.complex64)


class _Sc16File(BatchFile[npt.NDArray[np.complex64]]):
    def read(self) -> npt.NDArray[np.complex64]:
        """Read double-precision complex, interleaved I/Q samples in the binary format. A promotional
        conversion is applied.

        :return: 64-bit complex IQ samples.
        """
        data = np.fromfile(self.file_path, dtype=np.int16)
        return (data[0::2] + 1j * data[1::2]).astype(np.complex64)


@dataclasses.dataclass
class IQMetadata:
    """Stores metadata produced by the batched file sink block.

    :ivar center_frequencies: Center frequencies for each I/Q sample.
    :ivar num_samples: The number of I/Q samples at each center frequency.
    """

    center_frequencies: npt.NDArray[np.float32]
    num_samples: npt.NDArray[np.int32]


class _HdrFile(BatchFile[IQMetadata]):
    def read(self) -> IQMetadata:
        """Read metadata produced by the batched file sink block by parsing the interleaved
        center frequencies and the number of samples corresponding to each.

        :return: A container for the metadata
        """
        data = np.fromfile(self.file_path, dtype=np.float32)
        return IQMetadata(data[0::2], data[1::2].astype(np.int32))


class _FitsFile(BatchFile[spectre_core.spectrograms.Spectrogram]):
    def read(self) -> spectre_core.spectrograms.Spectrogram:
        """Read the FITS file and create a spectrogram."""
        with astropy.io.fits.open(self.file_path, mode="readonly") as hdulist:
            primary_hdu = hdulist["PRIMARY"]
            dynamic_spectra = primary_hdu.data
            bunit = primary_hdu.header["BUNIT"]

            date_obs = primary_hdu.header["DATE-OBS"]
            time_obs = primary_hdu.header["TIME-OBS"]
            spectrogram_start_datetime = datetime.datetime.strptime(
                f"{date_obs}T{time_obs}Z",
                spectre_core.config.TimeFormat.DATETIME,
            )

            bintable_hdu = hdulist[1]
            times = bintable_hdu.data["TIME"][0]
            frequencies = bintable_hdu.data["FREQUENCY"][0] * 1e6  # Convert to Hz

        # bunit is interpreted as a SpectrumUnit.
        spectrum_unit = spectre_core.spectrograms.SpectrumUnit(bunit)
        return spectre_core.spectrograms.Spectrogram(
            dynamic_spectra,
            times,
            frequencies,
            spectrum_unit,
            spectrogram_start_datetime,
        )


class IQStreamBatch(Base):

    def __init__(self, batches_dir_path: str, start_time: str, tag: str) -> None:
        """A batch of data derived from a stream of IQ samples from some receiver.

        Supports the following extensions:
        - `.fits`
        - `.fc32`
        - `.fc64`
        - `.sc8`
        - `.sc16`
        - `.hdr`

        :param start_time: The start time of the batch.
        :param tag: The batch name tag.
        """
        super().__init__(batches_dir_path, start_time, tag)

        self.add_file(_FitsFile, IQStreamBatchExtension.FITS)
        self.add_file(_Fc32File, IQStreamBatchExtension.FC32)
        self.add_file(_Fc64File, IQStreamBatchExtension.FC64)
        self.add_file(_Sc8File, IQStreamBatchExtension.SC8)
        self.add_file(_Sc16File, IQStreamBatchExtension.SC16)
        self.add_file(_HdrFile, IQStreamBatchExtension.HDR)

    @property
    def fits_file(self) -> _FitsFile:
        """The batch file corresponding to the `.fits` extension."""
        return typing.cast(_FitsFile, self.get_file(IQStreamBatchExtension.FITS))

    @property
    def fc32_file(self) -> _Fc32File:
        """The batch file corresponding to the `.fc32` extension."""
        return typing.cast(_Fc32File, self.get_file(IQStreamBatchExtension.FC32))

    @property
    def fc64_file(self) -> _Fc64File:
        """The batch file corresponding to the `.fc64` extension."""
        return typing.cast(_Fc64File, self.get_file(IQStreamBatchExtension.FC64))

    @property
    def sc8_file(self) -> _Sc8File:
        """The batch file corresponding to the `.sc8` extension."""
        return typing.cast(_Sc8File, self.get_file(IQStreamBatchExtension.SC8))

    @property
    def sc16_file(self) -> _Sc16File:
        """The batch file corresponding to the `.sc16` extension."""
        return typing.cast(_Sc16File, self.get_file(IQStreamBatchExtension.SC16))

    @property
    def hdr_file(self) -> _HdrFile:
        """The batch file corresponding to the `.hdr` extension."""
        return typing.cast(_HdrFile, self.get_file(IQStreamBatchExtension.HDR))

    @property
    def spectrogram_file(self) -> _FitsFile:
        return self.fits_file

    def read_iq(self, extension: str) -> npt.NDArray[np.complex64]:
        """Read I/Q samples from the batch."""
        if extension == IQStreamBatchExtension.FC32:
            return self.fc32_file.read()
        elif extension == IQStreamBatchExtension.FC64:
            return self.fc64_file.read()
        elif extension == IQStreamBatchExtension.SC8:
            return self.sc8_file.read()
        elif extension == IQStreamBatchExtension.SC16:
            return self.sc16_file.read()
        else:
            raise ValueError(f"Unsupported output type: {extension}")

    def cached_read_iq(self, extension: str) -> npt.NDArray[np.complex64]:
        """Read I/Q samples from the batch."""
        if extension == IQStreamBatchExtension.FC32:
            return self.fc32_file.cached_read()
        elif extension == IQStreamBatchExtension.FC64:
            return self.fc64_file.cached_read()
        elif extension == IQStreamBatchExtension.SC8:
            return self.sc8_file.cached_read()
        elif extension == IQStreamBatchExtension.SC16:
            return self.sc16_file.cached_read()
        else:
            raise ValueError(f"Unsupported output type: {extension}")

    def delete_iq(self, extension: str) -> None:
        """Delete I/Q samples from the batch."""
        if extension == IQStreamBatchExtension.FC32:
            self.fc32_file.delete()
        elif extension == IQStreamBatchExtension.FC64:
            self.fc64_file.delete()
        elif extension == IQStreamBatchExtension.SC8:
            self.sc8_file.delete()
        elif extension == IQStreamBatchExtension.SC16:
            self.sc16_file.delete()
        else:
            raise ValueError(f"Unsupported output type: {extension}")



# SPDX-FileCopyrightText: © 2024-2026 Jimmy Fitzpatrick <jcfitzpatrick12@gmail.com>
# This file is part of SPECTRE
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from ._validators import validate_one_of

EXPECTED_OUTPUT_TYPES: list[str] = ["fc32", "sc16", "sc8"]


def validate_output_type(output_type: str):
    """Checks the output type is supported."""
    validate_one_of(output_type, EXPECTED_OUTPUT_TYPES, "output_type")
